:- use_module(library(lists)).
:- use_module(library(between)).
:- use_module(library(clpb)).


test_mines(Rows, Cols, Mines, Bd, All) -->
	{length(Bd, Rows),
	Bd = [FirstRow|RestRows],
	length(FirstRow, Cols),
	maplist(same_length(FirstRow), RestRows),
	append(Bd, All)},
	[card([Mines], All)],
	play_mine1(Bd, []).

play_mine1(Bd, Asked) -->
	{select_field(Bd, Asked, R, C, E), !,
	write(Asked-R-C-E), nl,
	Ans = 0'2},
	(   process_ans(Ans, E, R, C, Bd)
	->  play_mine1(Bd, [R-C|Asked])
	).
play_mine1(_Bd, _Asked) --> [].

select_field(Bd, Asked, R, C, E) :-
	nth1(R, Bd, L), 
	nth1(C, L, E),		% E is in row R col C of Bd
	nonmember(R-C, Asked),  % R-C has not been clicked on yet
	E == 0,                 % E is false
	neighbs(n(R,C,Bd), Ns),
	\+ ground(Ns),          % R-C has a neighbour which is a variable
	!.
select_field(Bd, _Asked, R, C, E) :-
	nth1(R, Bd, L), nth1(C, L, E),
	var(E), !.

process_ans(Ans, 0, R, C, Bd) -->
	{Ans >= 0'0, Ans =< 0'8, Cnt is Ans - 0'0,
	neighbs(n(R, C, Bd), Ns)},
	[card([Cnt], Ns)].


neighbs(RCB, N7) :-
	findall(R-C, (   between(-1, 1, R), 
			 between(-1, 1, C),
			 \+ (R==0,C==0)
		     ), Offsets),
	scanlist(neighbour(RCB), Offsets, [], N7).


neighbour(n(R0, C0, Bd), ROf-COf, Nbs, [E|Nbs]) :-
	R is R0+ROf, C is C0+COf, nth1(R, Bd, Row), nth1(C, Row, E), !.
neighbour(_, _, Nbs, Nbs).

